import json
import logging
import os
import shutil
from pathlib import Path
from typing import Any, Dict, List, Optional, Set, Union

from dvc_studio_client.env import STUDIO_TOKEN
from dvc_studio_client.post_live_metrics import post_live_metrics
from funcy import set_in
from ruamel.yaml.representer import RepresenterError

from . import env
from .dvc import (
    get_dvc_repo,
    get_dvc_stage_template,
    get_random_exp_name,
    make_checkpoint,
    make_dvcyaml,
    mark_dvclive_only_ended,
    mark_dvclive_only_started,
)
from .error import InvalidDataTypeError, InvalidParameterTypeError, InvalidPlotTypeError
from .plots import PLOT_TYPES, SKLEARN_PLOTS, Image, Metric, NumpyEncoder
from .report import BLANK_NOTEBOOK_REPORT, make_report
from .serialize import dump_json, dump_yaml, load_yaml
from .studio import get_studio_updates
from .utils import env2bool, inside_notebook, matplotlib_installed, open_file_in_browser

logging.basicConfig()
logger = logging.getLogger("dvclive")
logger.setLevel(os.getenv(env.DVCLIVE_LOGLEVEL, "INFO").upper())

ParamLike = Union[int, float, str, bool, List["ParamLike"], Dict[str, "ParamLike"]]
StrPath = Union[str, Path]


class Live:
    def __init__(
        self,
        dir: str = "dvclive",  # noqa pylint: disable=redefined-builtin
        resume: bool = False,
        report: Optional[str] = "auto",
        save_dvc_exp: bool = False,
        dvcyaml: bool = True,
    ):
        self.summary: Dict[str, Any] = {}

        self._dir: str = dir
        self._resume: bool = resume or env2bool(env.DVCLIVE_RESUME)
        self._save_dvc_exp: bool = save_dvc_exp
        self._step: Optional[int] = None
        self._metrics: Dict[str, Any] = {}
        self._images: Dict[str, Any] = {}
        self._params: Dict[str, Any] = {}
        self._plots: Dict[str, Any] = {}
        self._outs: Set[StrPath] = set()
        self._inside_with = False
        self._dvcyaml = dvcyaml

        os.makedirs(self.dir, exist_ok=True)

        self._report_mode: Optional[str] = report
        self._report_notebook = None
        self._init_report()

        if self._resume:
            self._init_resume()
        else:
            self._init_cleanup()

        self._baseline_rev: Optional[str] = None
        self._exp_name: Optional[str] = None
        self._experiment_rev: Optional[str] = None
        self._inside_dvc_exp: bool = False
        self._dvc_repo = None
        self._include_untracked: List[str] = []
        self._init_dvc()

        self._latest_studio_step = self.step if resume else -1
        self._studio_events_to_skip: Set[str] = set()
        self._init_studio()

    def _init_resume(self):
        self._read_params()
        self._step = self.read_step()
        if self._step != 0:
            logger.info(f"Resuming from step {self._step}")
            self._step += 1
        logger.debug(f"{self._step=}")

    def _init_cleanup(self):
        for plot_type in PLOT_TYPES:
            shutil.rmtree(
                Path(self.plots_dir) / plot_type.subfolder, ignore_errors=True
            )

        for f in (
            self.metrics_file,
            self.report_file,
            self.params_file,
        ):
            if f and os.path.exists(f):
                os.remove(f)

    def _init_dvc(self):
        if os.getenv(env.DVC_EXP_BASELINE_REV, None):
            # `dvc exp` execution
            self._baseline_rev = os.getenv(env.DVC_EXP_BASELINE_REV, "")
            self._exp_name = os.getenv(env.DVC_EXP_NAME, "")
            self._inside_dvc_exp = True
            if self._save_dvc_exp:
                logger.warning(
                    "Ignoring `_save_dvc_exp` because `dvc exp run` is running"
                )
                self._save_dvc_exp = False

        self._dvc_repo = get_dvc_repo()
        if self._dvc_repo is None:
            if self._save_dvc_exp:
                logger.warning(
                    "Can't save experiment without a DVC Repo."
                    "\nYou can create a DVC Repo by calling `dvc init`."
                )
                self._save_dvc_exp = False
            return

        if self._inside_dvc_exp:
            return

        self._baseline_rev = self._dvc_repo.scm.get_rev()
        if self._save_dvc_exp:
            self._exp_name = get_random_exp_name(self._dvc_repo.scm, self._baseline_rev)
            mark_dvclive_only_started()
            self._include_untracked.append(self.dir)

    def _init_studio(self):
        if not os.getenv(STUDIO_TOKEN, None):
            logger.debug("Missing env var `STUDIO_TOKEN`, skipping `studio` report.")
            self._studio_events_to_skip.add("start")
            self._studio_events_to_skip.add("data")
            self._studio_events_to_skip.add("done")
        elif self._inside_dvc_exp:
            logger.debug("Skipping `studio` report `start` and `done` events.")
            self._studio_events_to_skip.add("start")
            self._studio_events_to_skip.add("done")
        elif self._dvc_repo is None:
            logger.warning(
                "Can't connect to Studio without a DVC Repo."
                "\nYou can create a DVC Repo by calling `dvc init`."
            )
            self._studio_events_to_skip.add("start")
            self._studio_events_to_skip.add("data")
            self._studio_events_to_skip.add("done")
        elif not self._save_dvc_exp:
            logger.warning(
                "Can't connect to Studio without creating a DVC experiment."
                "\nIf you have a DVC Pipeline, run it with `dvc exp run`."
                "\nIf you are using DVCLive alone, use `save_dvc_exp=True`."
            )
            self._studio_events_to_skip.add("start")
            self._studio_events_to_skip.add("data")
            self._studio_events_to_skip.add("done")
        else:
            response = post_live_metrics(
                "start", self._baseline_rev, self._exp_name, "dvclive"
            )
            if not response:
                logger.debug(
                    "`studio` report `start` event failed. "
                    "`studio` report cancelled."
                )
                self._studio_events_to_skip.add("start")
                self._studio_events_to_skip.add("data")
                self._studio_events_to_skip.add("done")

    def _init_report(self):
        if self._report_mode == "auto":
            if env2bool("CI") and matplotlib_installed():
                self._report_mode = "md"
            else:
                self._report_mode = "html"
        elif self._report_mode == "notebook":
            if inside_notebook():
                from IPython.display import HTML, display

                self._report_mode = "notebook"
                self._report_notebook = display(
                    HTML(BLANK_NOTEBOOK_REPORT), display_id=True
                )
            else:
                self._report_mode = "html"
        elif self._report_mode not in {None, "html", "notebook", "md"}:
            raise ValueError(
                "`report` can only be `None`, `auto`, `html`, `notebook` or `md`"
            )
        logger.debug(f"{self._report_mode=}")

    @property
    def dir(self) -> str:
        return self._dir

    @property
    def params_file(self) -> str:
        return os.path.join(self.dir, "params.yaml")

    @property
    def metrics_file(self) -> str:
        return os.path.join(self.dir, "metrics.json")

    @property
    def dvc_file(self) -> str:
        return os.path.join(self.dir, "dvc.yaml")

    @property
    def plots_dir(self) -> str:
        return os.path.join(self.dir, "plots")

    @property
    def report_file(self) -> Optional[str]:
        if self._report_mode in ("html", "md", "notebook"):
            if self._report_mode == "notebook":
                suffix = "html"
            else:
                suffix = self._report_mode
            return os.path.join(self.dir, f"report.{suffix}")
        return None

    @property
    def step(self) -> int:
        return self._step or 0

    @step.setter
    def step(self, value: int) -> None:
        self._step = value
        logger.debug(f"Step: {self.step}")

    def next_step(self):
        if self._step is None:
            self._step = 0

        self.make_summary()

        if self._dvcyaml:
            self.make_dvcyaml()

        self.make_report()
        self.make_checkpoint()
        self.step += 1

    def log_metric(self, name: str, val: Union[int, float], timestamp: bool = False):
        if not Metric.could_log(val):
            raise InvalidDataTypeError(name, type(val))

        if name in self._metrics:
            data = self._metrics[name]
        else:
            data = Metric(name, self.plots_dir)
            self._metrics[name] = data

        data.step = self.step
        data.dump(val, timestamp=timestamp)

        self.summary = set_in(self.summary, data.summary_keys, val)
        logger.debug(f"Logged {name}: {val}")

    def log_image(self, name: str, val):
        if not Image.could_log(val):
            raise InvalidDataTypeError(name, type(val))

        if name in self._images:
            data = self._images[name]
        else:
            data = Image(name, self.plots_dir)
            self._images[name] = data

        data.step = self.step
        data.dump(val)
        logger.debug(f"Logged {name}: {val}")

    def log_sklearn_plot(self, kind, labels, predictions, name=None, **kwargs):
        val = (labels, predictions)

        name = name or kind
        if name in self._plots:
            data = self._plots[name]
        elif kind in SKLEARN_PLOTS and SKLEARN_PLOTS[kind].could_log(val):
            data = SKLEARN_PLOTS[kind](name, self.plots_dir)
            self._plots[data.name] = data
        else:
            raise InvalidPlotTypeError(name)

        data.step = self.step
        data.dump(val, **kwargs)
        logger.debug(f"Logged {name}")

    def _read_params(self):
        if os.path.isfile(self.params_file):
            params = load_yaml(self.params_file)
            self._params.update(params)

    def _dump_params(self):
        try:
            dump_yaml(self._params, self.params_file)
        except RepresenterError as exc:
            raise InvalidParameterTypeError(exc.args) from exc

    def log_params(self, params: Dict[str, ParamLike]):
        """Saves the given set of parameters (dict) to yaml"""
        self._params.update(params)
        self._dump_params()
        logger.debug(f"Logged {params} parameters to {self.params_file}")

    def log_param(self, name: str, val: ParamLike):
        """Saves the given parameter value to yaml"""
        self.log_params({name: val})

    def log_artifact(self, path: StrPath):
        """Tracks a local file or directory with DVC"""
        if not isinstance(path, (str, Path)):
            raise InvalidDataTypeError(path, type(path))

        if self._dvc_repo is not None:
            try:
                stage = self._dvc_repo.add(path)
            except Exception as e:  # pylint: disable=broad-except
                logger.warning(f"Failed to dvc add {path}: {e}")
                return

            self._outs.add(path)
            dvc_file = stage[0].addressing

            if self._save_dvc_exp:
                self._include_untracked.append(dvc_file)
                self._include_untracked.append(
                    str(Path(dvc_file).parent / ".gitignore")
                )

    def make_summary(self, update_step: bool = True):
        if self._step is not None and update_step:
            self.summary["step"] = self.step
        dump_json(self.summary, self.metrics_file, cls=NumpyEncoder)

    def make_report(self):
        if "data" not in self._studio_events_to_skip:
            response = False
            if post_live_metrics is not None:
                metrics, params, plots = get_studio_updates(self)
                response = post_live_metrics(
                    "data",
                    self._baseline_rev,
                    self._exp_name,
                    "dvclive",
                    step=self.step,
                    metrics=metrics,
                    params=params,
                    plots=plots,
                )
            if not response:
                logger.warning(
                    "`post_to_studio` `data` event failed."
                    " Data will be resent on next call."
                )
            else:
                self._latest_studio_step = self.step

        if self._report_mode is not None:
            make_report(self)
            if self._report_mode == "html" and env2bool(env.DVCLIVE_OPEN):
                open_file_in_browser(self.report_file)

    def make_dvcyaml(self):
        make_dvcyaml(self)

    def end(self):
        if self._inside_with:
            # Prevent `live.end` calls inside context manager
            return
        self.make_summary(update_step=False)
        if self._dvcyaml:
            self.make_dvcyaml()

        if self._inside_dvc_exp and self._dvc_repo:
            self._dvc_repo.scm.add(self._dir)

        if "done" not in self._studio_events_to_skip:
            response = False
            if post_live_metrics is not None:
                kwargs = {}
                if self._experiment_rev:
                    kwargs["experiment_rev"] = self._experiment_rev
                response = post_live_metrics(
                    "done",
                    self._baseline_rev,
                    self._exp_name,
                    "dvclive",
                    **kwargs,
                )
            if not response:
                logger.warning("`post_to_studio` `done` event failed.")
            self._studio_events_to_skip.add("done")
            self._studio_events_to_skip.add("data")
        else:
            self.make_report()

        if self._save_dvc_exp:
            from dvc.exceptions import DvcException

            try:
                self._experiment_rev = self._dvc_repo.experiments.save(
                    name=self._exp_name,
                    include_untracked=self._include_untracked,
                    force=True,
                )
            except DvcException as e:
                logger.warning(f"Failed to save experiment:\n{e}")
            finally:
                mark_dvclive_only_ended()

        if self._dvc_repo and not self._inside_dvc_exp:
            from dvc.exceptions import DvcException

            try:
                path = os.path.join(self._dvc_repo.root_dir, "dvc.yaml")
                yaml = get_dvc_stage_template(self)
                logger.info(f"To run with DVC, add this to {path}:\n{yaml}")
            except DvcException as e:
                logger.warning(f"Failed to print DVC stage template:\n{e}")

    def make_checkpoint(self):
        if env2bool(env.DVC_CHECKPOINT):
            make_checkpoint()

    def read_step(self):
        if Path(self.metrics_file).exists():
            latest = self.read_latest()
            return latest.get("step", 0)
        return 0

    def read_latest(self):
        with open(self.metrics_file, encoding="utf-8") as fobj:
            return json.load(fobj)

    def __enter__(self):
        self._inside_with = True
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._inside_with = False
        self.end()
